/* xz81, an X-based ZX81 emulator (based on Z81, xnc100em, and xz80).
 * Copyright (C) 1994 Ian Collier. xz81 changes (C) 1995-6 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Intrinsic.h>

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#endif

#ifdef HAS_UNAME
#include <sys/utsname.h>
#endif

#include "z80.h"

#define MAX_DISP_LEN 256

#include "xz81.icon"

#if SCALE>1
static unsigned long scaleup[256]; /* to hold table of scaled up bytes,
                                 e.g. 00110100B -> 0000111100110000B */
#endif

unsigned char keyports[9]={0xff,0xff,0xff,0xff, 0xff,0xff,0xff,0xff, 0xff};

#define BORDER_WIDTH	(20*SCALE)
int rrshm=2,rrnoshm=4,mitshm=1;

unsigned char mem[65536];
unsigned char *dfile;
unsigned char *memptr[4]={mem,mem+0x4000,mem+0x8000,mem+0xc000};
unsigned char *iptr=mem;
unsigned long tstates=0,tsmax=(1<<30);
int hires=0,hiscrn;
int topspeed=0,lowref=0;

int memattr[4]={0,2,1,1};

int screen_dirty;
int hsize=256*SCALE,vsize=192*SCALE;
volatile int interrupted=0;
int forceok=0;
int scrn_freq=2;
int booting=1;
int input_wait=0;

unsigned char chrmap_old[24*32],chrmap[24*32];
unsigned char himap_old[192*32],himap[192*32];

int refresh_screen=1;



void sighandler(a)
int a;
{
if(interrupted<2) interrupted=1;
}


void dontpanic(a)
int a;
{
closedown();
exit(1);
}


char *libdir(char *file)
{
static char buf[1024];
sprintf(buf,"%s/%s",LIBDIR,file);
return(buf);
}



main(argc,argv)
int argc;
char **argv;
{
struct sigaction sa;
struct itimerval itv;
int tmp=1000/50;	/* 50 ints/sec */

// Load the ROM into memory
extern unsigned char zx81rom[];
memcpy(mem, zx81rom, 8192);
memcpy(&mem[8192], zx81rom, 8192);

hacks();
memset(mem+16384,0xff,49152);
memset(chrmap_old,0xff,768);

startup(&argc,argv);

if(argc==2) scrn_freq=atoi(argv[1]);
if(scrn_freq<1) scrn_freq=1;
if(scrn_freq>50) scrn_freq=50;

sa.sa_handler=dontpanic;
//sa.sa_mask=0;
memset(&sa.sa_mask, 0, sizeof(sa.sa_mask));
sa.sa_flags=SA_ONESHOT;

sigaction(SIGINT, &sa,NULL);
sigaction(SIGHUP, &sa,NULL);
sigaction(SIGILL, &sa,NULL);
sigaction(SIGTERM,&sa,NULL);
sigaction(SIGQUIT,&sa,NULL);
sigaction(SIGSEGV,&sa,NULL);

sa.sa_handler=sighandler;
memset(&sa.sa_mask, 0, sizeof(sa.sa_mask));
//sa.sa_mask=0;
sa.sa_flags=SA_RESTART;

sigaction(SIGALRM,&sa,NULL);

itv.it_interval.tv_sec=  tmp/1000;
itv.it_interval.tv_usec=(tmp%1000)*1000;
itv.it_value.tv_sec=itv.it_interval.tv_sec;
itv.it_value.tv_usec=itv.it_interval.tv_usec;
setitimer(ITIMER_REAL,&itv,NULL);

mainloop();

/* doesn't ever get here really, but all the same... */
closedown();
exit(0);
}


hacks()
{
/* remove fast mode stuff which screws it all up */
mem[0x4cc]=mem[0x4cd]=mem[0x4ce]=0;

/* replace kybd routine itself, with ld hl,(lastk):ret */
mem[0x2bb]=0x2a; mem[0x2bc]=0x25; mem[0x2bd]=0x40; mem[0x2be]=0xc9;

/* a 'force F/0' at 66h; ok, as we simulate NMI effects by other means */
mem[0x66]=0xcf; mem[0x67]=0x0e;

/* patch save routine */
mem[0x2fc]=0xed; mem[0x2fd]=0xfd;
mem[0x2fe]=0xc3; mem[0x2ff]=0x07; mem[0x300]=0x02;

/* patch load routine */
mem[0x347]=0xeb;
mem[0x348]=0xed; mem[0x349]=0xfc;
mem[0x34a]=0xc3; mem[0x34b]=0x07; mem[0x34c]=0x02;

/* make fast/slow more traceable (directly modify bit 7 rather than
 * bit 6 of CDFLAGS)
 */
mem[0xf29]=0xbe;
mem[0xf2e]=0xfe;

/* do 'out (0),a' when waiting for input (for fast mode display check) */
mem[0x4ca]=0xd3; mem[0x4cb]=0;			/* out (0),a */
mem[0x4cc]=0xcb; mem[0x4cd]=0x46;		/* loop: bit 0,(hl) */
mem[0x4ce]=0x28; mem[0x4cf]=0xfc;		/* jr z,loop */
mem[0x4d0]=0xd3; mem[0x4d1]=1;			/* out (1),a */
mem[0x4d2]=0;	/* nop, as filler */

/* skip some more weird fast mode stuff */
mem[0x2ec]=0xc9;

/* we need the following crock to support 'PAUSE'... */

mem[0xf3b]=0x29;	/* call 0229h */

mem[0x229]=0x22;	/* ld (04034h),hl:nop */
mem[0x22c]=0;

mem[0x23e]=0x2a; mem[0x23f]=0x25; mem[0x240]=0x40;	/* ld hl,(0403bh) */
mem[0x241]=0x23;					/* inc hl */
mem[0x242]=0x3e; mem[0x243]=0x7f;			/* ld a,07fh */
mem[0x244]=0xa4;					/* and h */
mem[0x245]=0xb5;					/* or l */
mem[0x246]=0xc0;					/* ret nz */

mem[0x247]=0x2a;	/* ld hl,(04034h) */
mem[0x248]=0x34;
mem[0x249]=0x40;

mem[0x24a]=0xc3;	/* jp 022dh */
mem[0x24b]=0x2d;
mem[0x24c]=0x02;

/* nop out ld (04034h),hl */
mem[0x23a]=0; mem[0x23a]=0; mem[0x23a]=0;

/* nop out call to 207h */
mem[0xf41]=0; mem[0xf42]=0; mem[0xf43]=0;
}


unsigned int in(h,l)
int h,l;
{
if(l==0xfe)	/* keyboard */
  switch(h)
    {
    case 0xfe:	return(keyports[0]);
    case 0xfd:	return(keyports[1]);
    case 0xfb:	return(keyports[2]);
    case 0xf7:	return(keyports[3]);
    case 0xef:	return(keyports[4]);
    case 0xdf:	return(keyports[5]);
    case 0xbf:	return(keyports[6]);
    case 0x7f:	return(keyports[7]);
    default:	return(255);
    }

fprintf(stderr," in l=%02X\n",l);
return(255);
}


unsigned int out(h,l,a)
int h,l,a;
{
if(l==0xfd) store(16443,128|fetch(16443));	/* i.e. set slow mode */
if(l==0x00) input_wait=1;	/* waiting for input */
if(l==0x01) input_wait=0;	/* not waiting for input :-) */

return(0);
}


hireschk(ix)
int ix;
{
refresh_screen=1;

if(ix<0x4000)
  hires=0;
else
  {
  int f,g,h,v;
  
  /* now try to find the hi-res screen.
   * we do this by looking for somewhere in ram where there
   * are 192 suitably-spaced EOL chars. The EOL char could theoretically be
   * a few different things, so we just test that it's the same.
   * this is SLOW, but it doesn't matter; this'll only need to
   * happen once, when the hi-res screen is set up.
   *
   * (um, ok, it's a little different now; I just kept hacking until
   * it worked with all three hi-res programs I tried it on :-))
   */
   
  for(f=0x8000-6400;f>=0x4000;f--)
    {
    v=mem[f+32];
    if((v&64)==0) continue;
    hires=1;
    for(g=0;g<192 && hires;g++)
      {
      if(mem[f+33*g]&64) hires=0;
      if(mem[f+32+33*g]!=v) hires=0;
      }
    if(hires)
      {
      hiscrn=f;
      return(0);
      }
    }
  }
}


hireschki(i)
int i;
{
iptr=mem+(i<<8);
}


update_kybd()
{
int y,b;
int lastk0=0,lastk1=0;	/* values for lastk and lastk+1 */
static int oldlk0=0,oldlk1=0;
int f;
unsigned short tmp;

/* bump frames */
f=(fetch2(16436)&32767);
if(f>0) f--;
store2(16436,(fetch2(16436)&32768)|f);

/* set max T-states per 1/50th to get the right speed */
if(topspeed)
  tsmax=(1<<30);
else
  if(fetch(16443)&128)
    {
    /* slow mode - 0.8 MHz, 16000 T-states per 1/50th */
    tsmax=16000;
    }
  else
    {
    /* fast mode - 3.2 MHz, 64000 T-states per 1/50th */
    tsmax=64000;
    }

/* fix top bits of key ports, and generate lastk */

for(y=0;y<8;y++)		/* 8 half-rows */
  {
  b=keyports[y];
  
  /* contribute to lastk if key was pressed */
  if((b&31)!=31)
    {
    /* set y bit in lastk0 if not shift */
    if(!(y==0 && ((b&31)|1)==31)) lastk0|=(1<<y);
    
    /* now set pos-in-half-row bit(s) in lastk1 */
    b=(((~b)&31)<<1);
    /* move bit 1 of b back down to bit 0 if it's shift bit */
    if(y==0) b=((b&0xfc)|((b&2)>>1));
  
    lastk1|=b;
    }
  }

/* see if this is different to prev. lastk[01] */
store(16443,(fetch(16443)&0xfe)|
	((((lastk1&0xfe)|lastk0) && (oldlk0!=lastk0 || oldlk1!=lastk1))?1:0));

oldlk0=lastk0; oldlk1=lastk1;

/* fix lastk values */
lastk0^=255; lastk1^=255;

/* put lastk stuff in place */
store(16421,lastk0);
store(16422,lastk1);

/* and FWIW, set MARGIN */
store(16424,55);
}


/* the ZX81 char is used to index into this, to give the ascii.
 * awkward chars are mapped to '_' (underscore), and the ZX81's
 * all-caps is converted to lowercase.
 * WARNING: this only covers 0<=char<=63!
 */
char zx2ascii[64]={
/* 000-009 */ ' ', '_', '_', '_', '_', '_', '_', '_', '_', '_', 
/* 010-019 */ '_', '\'','#', '$', ':', '?', '(', ')', '>', '<', 
/* 020-029 */ '=', '+', '-', '*', '/', ';', ',', '.', '0', '1', 
/* 030-039 */ '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 
/* 040-049 */ 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 
/* 050-059 */ 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 
/* 060-063 */ 'w', 'x', 'y', 'z'
};


save_p(a)
int a;
{
static char fname[128];
unsigned char *ptr=mem+a,*dptr=fname;
FILE *out;

/* so the filename is at ptr, in the ZX81 char set, with the last char
 * having bit 7 set. First, get that name in ASCII.
 */

memset(fname,0,sizeof(fname));
do
  *dptr++=zx2ascii[(*ptr)&63];
while((*ptr++)<127);

/* add '.p' */
strcat(fname,".p");

/* now open the file */
if((out=fopen(fname,"wb"))==NULL)
  {
  forceok=1;
  return(0);
  }

/* work out how much to write, and write it.
 * we need to write from 0x4009 to E_LINE inclusive.
 */
fwrite(mem+0x4009,1,fetch2(16404)-0x4009+1,out);

fclose(out);
}


load_p(a)
int a;
{
static char fname[128];
unsigned char *ptr=mem+(a&32767),*dptr=fname;
FILE *in;

if(a&32768) 	/* they did LOAD "" */
  {
  /* do an ls. This is crude, but so what? :-) */
  system("ls *.p");
  
  forceok=1;
  return(0);
  }

/* so the filename is at ptr, in the ZX81 char set, with the last char
 * having bit 7 set. First, get that name in ASCII.
 */

/* test for Xtender-style LOAD " STOP " to quit */
if(*ptr==227) dontpanic();

memset(fname,0,sizeof(fname));
do
  *dptr++=zx2ascii[(*ptr)&63];
while((*ptr++)<127);

/* add '.p' */
strcat(fname,".p");

/* now open the file */
if((in=fopen(fname,"rb"))==NULL)
  {
  forceok=1;
  return(0);
  }

/* just read it all */
fread(mem+0x4009,1,16384,in);

fclose(in);
}


fix_tstates()
{
tstates=0;
pause();
}


do_interrupt()
{
static int count=0;

if(booting)
  if(input_wait)
    booting=0;
  else
    return(0);

/* only do refresh() every 1/Nth */
count++;
if(count>=scrn_freq)
  count=0,refresh();
check_events(1);

/* be careful not to screw up any pending reset... */

if(interrupted==1)
  {
  update_kybd();
  interrupted=0;
  }
}



/* the remainder of xmain.c is based on xz80's xspectrum.c. */

#ifdef SunKludge
char *shmat();
char *getenv();
#endif

/* remember, this table is ignoring shifts... */
static struct {unsigned char port,mask;} keytable[]={
/*  SP      !        "        #        $        %        &        '  */
{7,0xfe},{8,0xff},{3,0xfd},{8,0xff},{5,0xf7},{8,0xff},{8,0xff},{8,0xff},
/*  (       )        *        +        ,        -        .        /  */
{5,0xfb},{5,0xfd},{4,0xfb},{6,0xfb},{7,0xfd},{6,0xf7},{7,0xfd},{0,0xef},
/*  0       1        2        3        4        5        6        7  */
{4,0xfe},{3,0xfe},{3,0xfd},{3,0xfb},{3,0xf7},{3,0xef},{4,0xef},{4,0xf7},
/*  8       9        :        ;        <        =        >        ?  */
{4,0xfb},{4,0xfd},{0,0xfd},{0,0xfb},{7,0xf7},{6,0xfb},{7,0xfb},{0,0xef},
/*  @       A        B        C        D        E        F        G  */
{8,0xff},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
/*  H       I        J        K        L        M        N        O  */
{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
/*  P       Q        R        S        T        U        V        W  */
{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
/*  X       Y        Z        [        \        ]        ^        _  */
{0,0xfb},{5,0xef},{0,0xfd},{8,0xff},{8,0xff},{8,0xff},{8,0xff},{8,0xff},
/*  `       a        b        c        d        e        f        g  */
{8,0xff},{1,0xfe},{7,0xef},{0,0xf7},{1,0xfb},{2,0xfb},{1,0xf7},{1,0xef},
/*  h       i        j        k        l        m        n        o  */
{6,0xef},{5,0xfb},{6,0xf7},{6,0xfb},{6,0xfd},{7,0xfb},{7,0xf7},{5,0xfd},
/*  p       q        r        s        t        u        v        w  */
{5,0xfe},{2,0xfe},{2,0xf7},{1,0xfd},{2,0xef},{5,0xf7},{0,0xef},{2,0xfd},
/*  x       y        z        {        |        }        ~       DEL */
{0,0xfb},{5,0xef},{0,0xfd},{8,0xff},{8,0xff},{8,0xff},{8,0xff},{8,0xff}
};


static Display *display;
static Screen *scrptr;
static int screen;
static Window root;
static Colormap cmap;
static GC maingc;
static GC fggc, bggc;
static Window borderwin, mainwin;
static XImage *ximage;
static unsigned char *image;
static int linelen;
static int black,white;
static int bytecycles;
static int framebytes;
#ifdef MITSHM
static XShmSegmentInfo xshminfo;
#endif
static int invert=0;
static int borderchange=1;

static int is_local_server(name)
char *name;
{
#ifdef HAS_UNAME
   struct utsname un;
#else
   char sysname[MAX_DISP_LEN];
#endif

   if(name[0]==':')return 1;
   if(!strncmp(name,"unix",4))return 1;
   if(!strncmp(name,"localhost",9))return 1;

#ifdef HAS_UNAME
   uname(&un);
   if(!strncmp(name,un.sysname,strlen(un.sysname)))return 1;
   if(!strncmp(name,un.nodename,strlen(un.nodename)))return 1;
#else
   gethostname(sysname,MAX_DISP_LEN);
   if(!strncmp(name,sysname,strlen(sysname)))return 1;
#endif
   return 0;
}


static Display *open_display(argc,argv)
int *argc;
char **argv;
{
   int i;
   char *ptr;
   char *myname="xz81",*myclass="Xz81";
   char dispname[MAX_DISP_LEN];
   Display *display;

   if((ptr=getenv("DISPLAY")))
     strcpy(dispname,ptr);
   else
     strcpy(dispname,":0.0");
   
   if(!(display=XOpenDisplay(dispname))){
      fprintf(stderr,"Unable to open display %s\n",dispname);
      exit(1);
   }

#ifdef MITSHM   
   mitshm=1;
#else
   mitshm=0;
#endif
   
   /* XXX deal with args here */

   if(mitshm && !is_local_server(dispname)){
      fputs("Disabling MIT-SHM on remote X server\n",stderr);
      mitshm=0;
   }
   return display;
}


static int image_init()
{
//printf("Image depth %d\n", DefaultDepth(display, screen));
#ifdef MITSHM
   if(mitshm){
      ximage=XShmCreateImage(display,DefaultVisual(display,screen),
             DefaultDepth(display,screen),ZPixmap,NULL,&xshminfo,
             hsize,vsize);
      if(!ximage){
         fputs("Couldn't create X image\n",stderr);
         return 1;
      }
      xshminfo.shmid=shmget(IPC_PRIVATE,
               ximage->bytes_per_line*(ximage->height+1),IPC_CREAT|0777);
      if(xshminfo.shmid == -1){
         perror("Couldn't perform shmget");
         return 1;
      }
      xshminfo.shmaddr=ximage->data=shmat(xshminfo.shmid,0,0);
      if(!xshminfo.shmaddr){
         perror("Couldn't perform shmat");
         return 1;
      }
      xshminfo.readOnly=0;
      if(!XShmAttach(display,&xshminfo)){
         perror("Couldn't perform XShmAttach");
         return 1;
      }
      scrn_freq=rrshm;
   } else
#endif
   {
      ximage=XCreateImage(display,DefaultVisual(display,screen),
             DefaultDepth(display,screen),ZPixmap,0,NULL,hsize,vsize,
             8,0);
      if(!ximage){
         perror("XCreateImage failed");
         return 1;
      }
      ximage->data=malloc(ximage->bytes_per_line*(ximage->height+1));
      if(!ximage->data){
         perror("Couldn't get memory for XImage data");
         return 1;
      }
      scrn_freq=rrnoshm;
   }
   // printf("%d bytes per line\n", ximage->bytes_per_line);
   linelen=ximage->bytes_per_line/SCALE;
   /**
   if(linelen!=32 && linelen!=256)
      fprintf(stderr,"Line length=%d; expect strange results!\n",linelen);
   **/
   
#if 0
   if(linelen==32 &&
         (BitmapBitOrder(display)!=MSBFirst || ImageByteOrder(display)!=MSBFirst))
      fprintf(stderr,"BitmapBitOrder=%s and ImageByteOrder=%s.\n",
         BitmapBitOrder(display)==MSBFirst?"MSBFirst":"LSBFirst",
         ImageByteOrder(display)==MSBFirst?"MSBFirst":"LSBFirst"),
      fputs("If the display is mixed up, please mail me these values\n",stderr),
      fputs("and describe the display as accurately as possible.\n",stderr);
#endif
   ximage->data+=linelen;
   image=ximage->data;
   return 0;
}


static void notify(argc,argv)
int *argc;
char **argv;
{
   Pixmap icon;
   XWMHints xwmh;
   XSizeHints xsh;
   XClassHint xch;
   XTextProperty appname, iconname;
   char *apptext;
   char *icontext="xz81";

   icon=XCreatePixmapFromBitmapData(display,root,icon_bits,
        icon_width,icon_height,black,white,DefaultDepth(display,screen));
   apptext="xz81";
   xsh.flags=PSize|PMinSize|PMaxSize;
   xsh.min_width=hsize;
   xsh.min_height=vsize;
   xsh.max_width=hsize+BORDER_WIDTH*2;
   xsh.max_height=vsize+BORDER_WIDTH*2;
   if(!XStringListToTextProperty(&apptext,1,&appname)){
      fputs("Can't create a TextProperty!",stderr);
      return;
   }
   if(!XStringListToTextProperty(&icontext,1,&iconname)){
      fputs("Can't create a TextProperty!",stderr);
      return;
   }
   xwmh.initial_state=NormalState;
   xwmh.input=1;
   xwmh.icon_pixmap=icon;
   xwmh.flags=StateHint|IconPixmapHint|InputHint;
   xch.res_name="xz81";
   xch.res_class="Xz81";
   XSetWMProperties(display,borderwin,&appname,&iconname,argv,
      *argc,&xsh,&xwmh,&xch);
}


#if SCALE>1
static void scaleup_init(){
   int j, k, l, m;
   unsigned long bigval; /* SCALING must be <= 4! */
   for(l=0;l<256;scaleup[l++]=bigval)
      for(m=l,bigval=j=0;j<8;j++) {
         for(k=0;k<SCALE;k++)
            bigval=(bigval>>1)|((m&1)<<31);
         m>>=1;
      }
}
#endif


startup(argc,argv)
int *argc;
char **argv;
{
   display=open_display(argc,argv);
   if(!display){
      fputs("Failed to open X display\n",stderr);
      exit(1);
   }
   invert=0;
   screen=DefaultScreen(display);
   scrptr=DefaultScreenOfDisplay(display);
   root=DefaultRootWindow(display);
   white=WhitePixel(display,screen);
   black=BlackPixel(display,screen);
   maingc=XCreateGC(display,root,0,NULL);
   XCopyGC(display,DefaultGC(display,screen),~0,maingc);
   XSetGraphicsExposures(display,maingc,0);
   fggc=XCreateGC(display,root,0,NULL);
   XCopyGC(display,DefaultGC(display,screen),~0,fggc);
   XSetGraphicsExposures(display,fggc,0);
   bggc=XCreateGC(display,root,0,NULL);
   XCopyGC(display,DefaultGC(display,screen),~0,bggc);
   XSetGraphicsExposures(display,bggc,0);
   cmap=DefaultColormap(display,screen);
   if(image_init()){
      if(mitshm){
         fputs("Failed to create X image - trying again with mitshm off\n",stderr);
         mitshm=0;
         if(image_init())exit(1);
      }
      else exit(1);
   }

#if SCALE>1
   if(linelen==32) scaleup_init();
#endif

   borderwin=XCreateSimpleWindow(display,root,0,0,
             hsize+BORDER_WIDTH*2,vsize+BORDER_WIDTH*2,0,0,0);
   mainwin=XCreateSimpleWindow(display,borderwin,BORDER_WIDTH,
             BORDER_WIDTH,hsize,vsize,0,0,0);
   notify(argc,argv);
   XSelectInput(display,borderwin,KeyPressMask|KeyReleaseMask|
      ExposureMask|EnterWindowMask|LeaveWindowMask|
      StructureNotifyMask);
   XMapRaised(display,borderwin);
   XMapRaised(display,mainwin);
   XFlush(display);
   refresh_screen=1;
}


static int process_keypress(kev)
XKeyEvent *kev;
{
   char buf[3];
   KeySym ks;

   XLookupString(kev,buf,2,&ks,NULL);

   switch(ks){
      case XK_F10:
        dontpanic();
        /* doesn't return */
        break;
      
      case XK_Escape:
        interrupted=2;	/* will cause a reset */
        memset(mem+16384,0xff,49152);
        iptr=mem;
        tsmax=(1<<30);
        hires=0; refresh_screen=1;
        booting=1; input_wait=0;
        break;
      
      case XK_Return: 
              keyports[6]&=0xfe; break;
      case XK_Control_L: case XK_Shift_L: case XK_Shift_R:
      case XK_Alt_L: case XK_Alt_R: case XK_Meta_L: case XK_Meta_R:
              keyports[0]&=0xfe; break;
      case XK_BackSpace: case XK_Delete:
      case XK_Tab:
      case XK_Up:
      case XK_Down:
      case XK_Left:
      case XK_Right:
              /* XXX unmapped */ break;
#if 0	/* XXX */
      case XK_minus:
              keyports[8]|=0x02; break;
      case XK_underscore:
              keyports[8]|=0x02; break;
      case XK_equal:
              keyports[7]|=0x01; break;
      case XK_plus:
              keyports[7]|=0x01; break;
      case XK_semicolon:
              keyports[9]|=0x10; break;
      case XK_colon:
              keyports[9]|=0x10; break;
      case XK_apostrophe:
              keyports[8]|=0x10; break;
      case XK_quotedbl:
              keyports[3]|=0x02; break;
      case XK_exclam:
              keyports[2]|=0x04; break;
      case XK_at:
              keyports[8]|=0x10; break;
      case XK_numbersign:
              keyports[6]|=0x10; break;
      case XK_dollar:
              keyports[4]|=0x01; break;
      case XK_percent:
              keyports[1]|=0x40; break;
      case XK_asciicircum:
              keyports[6]|=0x01; break;
      case XK_ampersand:
              keyports[7]|=0x02; break;
      case XK_asterisk:
              keyports[8]|=0x01; break;
      case XK_parenleft:
              keyports[9]|=0x02; break;
      case XK_parenright:
              keyports[9]|=0x01; break;
      case XK_comma:
              keyports[8]|=0x80; break;
      case XK_less:
              keyports[8]|=0x80; break;
      case XK_period:
              keyports[9]|=0x80; break;
      case XK_greater:
              keyports[9]|=0x80; break;
      case XK_slash:
              keyports[6]|=0x20; break;
      case XK_question:
              keyports[6]|=0x20; break;
      case XK_backslash: case XK_bar:
              keyports[7]|=0x04; break;
              break;
#endif
      default:
              if((int)(ks-=32)>=0 && ks<128)
                 keyports[keytable[ks].port]&=keytable[ks].mask;
   }
   return 0;
}


static void process_keyrelease(kev)
XKeyEvent *kev;
{
   char buf[3];
   KeySym ks;

   XLookupString(kev,buf,2,&ks,NULL);

   switch(ks){
      case XK_Return: 
              keyports[6]|=1; break;
      case XK_Control_L: case XK_Shift_L: case XK_Shift_R:
      case XK_Alt_L: case XK_Alt_R: case XK_Meta_L: case XK_Meta_R:
              keyports[0]|=1; break;
      case XK_BackSpace: case XK_Delete:
      case XK_Escape:
      case XK_Tab:
      case XK_Up:
      case XK_Down:
      case XK_Left:
      case XK_Right:
              /* XXX unmapped */ break;
#if 0	/* XXX */
      case XK_minus:
              keyports[8]&=~0x02; break;
      case XK_underscore:
              keyports[8]&=~0x02; break;
      case XK_equal:
              keyports[7]&=~0x01; break;
      case XK_plus:
              keyports[7]&=~0x01; break;
      case XK_semicolon:
              keyports[9]&=~0x10; break;
      case XK_colon:
              keyports[9]&=~0x10; break;
      case XK_apostrophe:
              keyports[8]&=~0x10; break;
      case XK_quotedbl:
              keyports[3]&=~0x02; break;
      case XK_exclam:
              keyports[2]&=~0x04; break;
      case XK_at:
              keyports[8]&=~0x10; break;
      case XK_numbersign:
              keyports[6]&=~0x10; break;
      case XK_dollar:
              keyports[4]&=~0x01; break;
      case XK_percent:
              keyports[1]&=~0x40; break;
      case XK_asciicircum:
              keyports[6]&=~0x01; break;
      case XK_ampersand:
              keyports[7]&=~0x02; break;
      case XK_asterisk:
              keyports[8]&=~0x01; break;
      case XK_parenleft:
              keyports[9]&=~0x02; break;
      case XK_parenright:
              keyports[9]&=~0x01; break;
      case XK_comma:
              keyports[8]&=~0x80; break;
      case XK_less:
              keyports[8]&=~0x80; break;
      case XK_period:
              keyports[9]&=~0x80; break;
      case XK_greater:
              keyports[9]&=~0x80; break;
      case XK_slash:
              keyports[6]&=~0x20; break;
      case XK_question:
              keyports[6]&=~0x20; break;
      case XK_backslash: case XK_bar:
              keyports[7]&=~0x04; break;
              break;
#endif
      default:
              if((int)(ks-=32)>=0 && ks<96)
                 keyports[keytable[ks].port]|=~keytable[ks].mask;
   }
   return;
}


int check_events(flag)
int flag;
{
   static int answer=0;
   int i;
   static XEvent xev;
   XConfigureEvent *conf_ev;
   XCrossingEvent *cev;
   
   while (XEventsQueued(display,QueuedAfterReading)){
      XNextEvent(display,&xev);
      switch(xev.type){
         case Expose: refresh_screen=1;
            break;
         case ConfigureNotify:
            conf_ev=(XConfigureEvent *)&xev;
            XMoveWindow(display,mainwin,(conf_ev->width-hsize)/2,
                        (conf_ev->height-vsize)/2);
            break;
         case MapNotify: case UnmapNotify: case ReparentNotify:
            break;
         case EnterNotify:
            cev=(XCrossingEvent *)&xev;
            if(cev->detail!=NotifyInferior)
               XAutoRepeatOff(display),XFlush(display);
            break;
         case LeaveNotify:
            cev=(XCrossingEvent *)&xev;
            if(cev->detail!=NotifyInferior)
               XAutoRepeatOn(display),XFlush(display);
            break;
         case KeyPress: i=process_keypress((XKeyEvent *)&xev);
            if(!answer)answer=i;
            break;
         case KeyRelease: process_keyrelease((XKeyEvent *)&xev);
            break;
         default:
            fprintf(stderr,"unhandled X event, type %d\n",xev.type);
      }
   }
   if(flag){
      i=answer;
      answer=0;
      return i;
   }
   else return 0;
}



/* XXX tryout */

/* to redraw the (low-res, at least) screen, we translate it into
 * a more normal char. map, then find the smallest rectangle which
 * covers all the changes, and update that.
 */

refresh(){
   int h,i,j,k,l,m,n;
   int attr;
   unsigned char *ptr,*cptr,*dst,*tmp;
   unsigned char val;
   int x,y,a,b,c,d,inv,x2,mask;
   int fast=(!(fetch(16443)&128) && !input_wait);
   int pasteol;
   int xmin,ymin,xmax,ymax;
   int ofs;
//printf("refresh()\n");     
   if(borderchange>0)
     {
     /* XXX what about expose events? need to set borderchange... */
     XSetWindowBackground(display,borderwin,white);
     XClearWindow(display,borderwin);
     XFlush(display);
     borderchange=0;
     }

   /* frustratingly, hi and lo-res modes are sufficiently different to
    * require separate code for each, but similar enough to mean that
    * some code has to be duplicated. :-( (And of course, speed concerns
    * `prevent' us from moving the duplicated code out into a separate
    * function.)
    */
   
   if(hires)
     {
printf("Draw hires\n");
     /* draw hi-res screen */
     ptr=mem+hiscrn;
     
     xmin=31; ymin=23; xmax=0; ymax=0;
     
     ofs=0;
     for(y=0;y<192;y++,ptr++)
       {
       for(x=0;x<32;x++,ptr++,ofs++)
         {
         c=*ptr;
         
         if((himap[ofs]=c)!=himap_old[ofs] || refresh_screen)
           {
           /* update size of area to be drawn */
           if(x<xmin) xmin=x;
           if(y/8<ymin) ymin=y/8;
           if(x>xmax) xmax=x;
           if(y/8>ymax) ymax=y/8;
           
           /* draw byte into image */
           inv=(c&128); c&=63;
           d=iptr[c*8]; if(inv) d^=255;
           
           if(linelen==32)
             {
             /* 1-bit mono */
             /* XXX doesn't support SCALE>1 */
             image[y*linelen+x]=~d;
             }
	     
           else // 8, 16, 24-bit
             {
	     int mask = 0;
	     int scale = 0;
	     int bytes_per_pixel = SCALE * linelen / 256;
	     for (scale = 0; scale < 2; ++scale)
	     {
	       int index = ((((y * 8) + b) * hsize) + (scale * (hsize / SCALE))
				+ (x * 8)) * bytes_per_pixel;
	       for (mask = 256; mask; mask >>= 1, index += bytes_per_pixel)
	 	     if (d & mask)
	     		memset(&image[index], 0, bytes_per_pixel);
		     else
		     	memset(&image[index], 0xff, bytes_per_pixel);
	     }
	     }
           }
         }
       }
     
     /* now, copy new to old for next time */
     memcpy(himap_old,himap,32*192);
     
     /* end of hi-res screen drawing */
     }
   else
     {
	//printf("Draw lo res\n");
     /* draw normal lo-res screen */
     
     /* translate to char map, comparing against previous */
     
     ptr=mem+fetch2(16396);	/* D_FILE */
     
     /* since we can't just do "don't bother if it's junk" as we always
      * need to draw a screen, just draw *valid* junk that won't result
      * in a segfault or anything. :-)
      */
     if(ptr-mem<0 || ptr-mem>0xf000) ptr=mem+0xf000;
     ptr++;	/* skip first HALT */
     
     cptr=mem+0x1e00;	/* char. set */
     
     xmin=31; ymin=23; xmax=0; ymax=0;
     
     ofs=0;
     for(y=0;y<24;y++,ptr++)
       {
       pasteol=0;
       for(x=0;x<32;x++,ptr++,ofs++)
         {
         if(pasteol || (*ptr)==0x76)
           {
           pasteol=1;
           c=0;		/* space */
           ptr--;
           }
         else
           c=*ptr;
refresh_screen=1;     
         if((chrmap[ofs]=c)!=chrmap_old[ofs] || refresh_screen)
           {
           /* update size of area to be drawn */
           if(x<xmin) xmin=x;
           if(y<ymin) ymin=y;
           if(x>xmax) xmax=x;
           if(y>ymax) ymax=y;
           
           /* draw character into image */
           inv=(c&128); c&=63;
           
           for(b=0;b<8;b++)
             {
             d=cptr[c*8+b]; if(inv) d^=255;
             
             if(linelen==32)
               {
               /* 1-bit mono */
               /* XXX doesn't support SCALE>1 */
               image[(y*8+b)*linelen+x]=~d;
               }
           else // 8, 16, 24-bit
	   {
	     int mask = 0;
	     int scale = 0;
	     int bytes_per_pixel = SCALE * linelen / 256;
	     for (scale = 0; scale < 2; ++scale)
	     {
	       int index = ((((y * 8) + b) * hsize) + (scale * (hsize / SCALE))
				+ (x * 8)) * bytes_per_pixel;
	       for (mask = 256; mask; mask >>= 1, index += bytes_per_pixel)
	 	     if (d & mask)
	     		memset(&image[index], 0, bytes_per_pixel);
		     else
		     	memset(&image[index], 0xff, bytes_per_pixel);
	     }
	   }


#if 0
             else
               {
               /* assume 8-bit */
               tmp=image+((y*8+b)*hsize+x*8)*SCALE;
               mask=256;
               while((mask>>=1))
#if SCALE<2
                 /* i.e. actual size */
                 *tmp++=(d&mask)?black:white;
#else
                 {
                 m=((d&mask)?black:white);
                 for(j=0;j<SCALE;j++)
                   for(k=0;k<SCALE;k++)
                     tmp[j*hsize+k]=m;
                 tmp+=SCALE;
                 }
#endif
               }
#endif


	       
             }
           }
         }
       }
     
     /* now, copy new to old for next time */
     memcpy(chrmap_old,chrmap,768);
     
     /* end of lo-res screen drawing */
     }

/* next bit happens for both hi and lo-res */

if(refresh_screen) xmin=0,ymin=0,xmax=31,ymax=23;

//if(xmax>=xmin && ymax>=ymin)
//  {     
#ifdef MITSHM
   if(mitshm)
     XShmPutImage(display,mainwin,maingc,ximage,
       xmin*8*SCALE,ymin*8*SCALE,xmin*8*SCALE,ymin*8*SCALE,
       (xmax-xmin+1)*8*SCALE,(ymax-ymin+1)*8*SCALE,0);
   else
#endif
     XPutImage(display,mainwin,maingc,ximage,
       xmin*8*SCALE,ymin*8*SCALE,xmin*8*SCALE,ymin*8*SCALE,
       (xmax-xmin+1)*8*SCALE,(ymax-ymin+1)*8*SCALE);
//  }

XFlush(display);
refresh_screen=0;
}


closedown(){
#ifdef MITSHM
   if(mitshm){
      XShmDetach(display,&xshminfo);
      XDestroyImage(ximage);
      shmdt(xshminfo.shmaddr);
      shmctl(xshminfo.shmid,IPC_RMID,0);
   }
#endif
   XAutoRepeatOn(display);
   XCloseDisplay(display);
}

